% Demo: Classifying MNIST handwritten digits using Gaussian NBCs

clear all;

% settings
full_cov = 1;

% load data
load('mnist_all.mat');

% put data into matrix and vector format
train_x = double(train0) / 255;
train_y = [0 * ones(size(train0, 1), 1)];
test_x = double(test0) / 255;
test_y = [0 * ones(size(test0, 1), 1)];
for i = 1:9,
    eval(sprintf('train_x = [train_x; double(train%d) / 255];', i));
    eval(sprintf('train_y = [train_y; %d * ones(size(train%d, 1), 1)];', i, i));
    eval(sprintf('test_x = [test_x; double(test%d) / 255];', i));
    eval(sprintf('test_y = [test_y; %d * ones(size(test%d, 1), 1)];', i, i));
end

%% show some examples from the test set
subplot(1, 10, 1);
suptitle('Examples from the training set');
indices = randi(size(train_x, 1),10,1);
for i = 1:10,
    subplot(1, 10, i);
    imshow(reshape(train_x(indices(i),:), 28, 28)')
    title(sprintf('#%d', train_y(indices(i))));
end
pause
clf

%% train model (estimate class conditional distribution, etc.)
close all
P_y = hist(train_y, 10) / size(train_y, 1); % estimate P(y)
histogram(train_y, 0.5 + 0:10);
title('label histogram');
pause
clf

%%
n = 28*28;
means = zeros(10, n);
if full_cov,
    vars = zeros(10, n, n);
else
    vars = zeros(10, n);
end
%loop over all lables y
for i = 1:10,
    %empirical mean
    means(i,:) = mean(train_x(train_y == (i-1),:)); 
    
    %full gaussian model
    if full_cov,
        %empirical variance
        vars(i, :, :) = cov(train_x(train_y == (i-1), :));
        vars(i,:,:) = squeeze(vars(i,:,:)) + 1e-3 * eye(28*28);
    %gaussian naive bayes model
    else
        %emprical variance
        vars(i,:) = var(train_x(train_y == (i-1), :));
        vars(i,:) = 1e-3 + vars(i,:);
    end
end

close all
subplot(1, 10, 1);
suptitle('Estimated means');
for i = 1:10,
    subplot(1, 10, i);
    m = means(i, :);
    imshow(reshape(m, 28, 28)');
    title(sprintf('#%d', i-1));
end
pause
clf

%% generate samples from distribution
subplot(5, 10, 1);
suptitle('Estimated means');
for k = 1:5,
    for i = 1:10,
        subplot(5, 10, i + (k - 1) * 10);
        class = i - 1; % randint(1, 1, 10);
        if full_cov,
            obs = mvnrnd(means(class+1,:), squeeze(vars(class+1,:,:)));
        else
            obs = mvnrnd(means(class+1,:), vars(class+1,:));
        end
        imshow(reshape(obs(:), 28, 28)');
        title(sprintf('#%d', class));
    end
end
pause
clf
close all

%% classify data
n_total = size(test_y, 1);
n_correct = 0;
probs_log = zeros(n_total, 10);
ct = log(0.99);

for i = 1:10,
    %log likelihood of feature given label i
    if full_cov,
        probs_log(:,i) = log_mvnpdf(test_x, means(i,:), squeeze(vars(i,:,:)));
    else
        probs_log(:,i) = log_mvnpdf(test_x, means(i,:), diag(vars(i,:)));
    end
    %log(p(y|x)) ~ log(p(x|y)p(y)
   
    probs_log(:,i) = probs_log(:,i) + log(P_y(i));
end


[a,b] = max(probs_log, [], 2);
n_correct = sum(b == (test_y + 1));
fprintf(1, 'Correct: %d/%d (%f %%)\n', n_correct, n_total, n_correct / n_total * 100);

