% Demo of active learning with L2-regularized logistic regression
% Author: Andreas Krause (ETHZ)

function logisticregression_active_demo
clf; clear;

lambda = .001; %regularization parameter
use_std = .2; %noise injected

%prepare data
gridsize = .02;

npc = 50; % points per class
n = 2*npc;
X(1:npc,:)=randn(npc,2)*use_std+1;
Y(1:npc)=1;
X((npc+1):n,:)=randn(npc,2)*use_std;
Y((npc+1):n)=-1;
minX = min(X(:,1))-.1; maxX = max(X(:,1))+.1;
minY = min(X(:,2))-.1; maxY = max(X(:,2))+.1;

T = 20;

w = zeros(3,1);
useind = [];

for t=1:T
    % some plotting

    % predict discriminant for plotting
    [Xg,Yg] = meshgrid(minX:gridsize:maxX,minY:gridsize:maxY);
    fval = ([Xg(:) Yg(:) ones(length(Xg(:)),1)]*w);
    pval = 1./(1+exp(-fval)); %probabilities on test grid

    clf
    axis([minX maxX minY maxY]);
    hold on
    plot(X(1:npc,1),X(1:npc,2),'o');
    plot(X((npc+1):n,1),X((npc+1):n,2),'s');
    plot(X(useind,1),X(useind,2),'ks','markersize',16);
    contour(Xg,Yg,reshape(pval,size(Xg)),linspace(0.0001,.9999,20));
    contour(Xg,Yg,reshape(sign(fval),size(Xg)));
    hold off
    
    uncertainty = -abs([X ones(n,1)]*w);
    uncertainty(useind)=-inf;
    [~,most_uncertain]=max(uncertainty);
    useind=[useind most_uncertain];
    w = train_logistic(X(useind,:),Y(useind),lambda);
    pause
% so
end

function w = train_logistic(X,Y,lambda)
[n,d] = size(X);
w = rand(d+1,1); %coefficients
niter = 1000;
for i = 1:niter
    j = randi(n,1); %pick random data point
        
    % Stochastic gradient descent step
    eta = min(5/i,1); %learning rate
    w = w+eta*Y(j)*[X(j,:)';1]/(1+exp(Y(j)*([X(j,:),1]*w)));
    w = w-eta*lambda*w; %gradient w.r.t. regularizer
    
end

