% Demo of "doubtful" logistic regression
% Author: Andreas Krause (ETHZ)

clf; clear;

lambda = .00001; %regularization parameter
use_std = .2; %noise injected
cost_ask = .5; %when is asking more expensive then misclassification; if .5, this is ordinary logistic regression

%prepare data
gridsize = .02;

npos = 50;  % number of positive examples
nneg = 50; % number of negative examples
n = npos+nneg;
X(1:npos,:)=randn(npos,2)*use_std+1;
Y(1:npos)=1;
X((npos+1):n,:)=randn(nneg,2)*use_std;
Y((npos+1):n)=-1;
minX = min(X(:,1))-.1; maxX = max(X(:,1))+.1;
minY = min(X(:,2))-.1; maxY = max(X(:,2))+.1;



w = rand(3,1); %coefficients
niter = 5000;
for i = 1:niter
    j = randi(n,1); %pick random data point
        
    % Stochastic gradient descent step
    eta = min(5/i,1); %learning rate
    w = w+eta*Y(j)*[X(j,:)';1]/(1+exp(Y(j)*([X(j,:),1]*w)));
    w = w-eta*lambda*w; %gradient w.r.t. regularizer
    
end
% some plotting

% predict discriminant for plotting
[Xg,Yg] = meshgrid(minX:gridsize:maxX,minY:gridsize:maxY);
fval = ([Xg(:) Yg(:) ones(length(Xg(:)),1)]*w);
pval = 1./(1+exp(-fval));

clf
axis([minX maxX minY maxY]);
hold on
plot(X(1:npos,1),X(1:npos,2),'o');
plot(X((npos+1):n,1),X((npos+1):n,2),'s');
contour(Xg,Yg,reshape(pval,size(Xg)),linspace(0.0001,.9999,20));
contour(Xg,Yg,reshape(sign(pval-cost_ask),size(Xg)));
contour(Xg,Yg,reshape(sign(pval-(1-cost_ask)),size(Xg)));
hold off
